CLI_STATUS_STARTLINE = 5
XML_HEADERS = {
    "Accept-Encoding": "gzip",
    "X-Requested-With": "XMLHttpRequest",
    "User-Agent": "Go-http-client/1.1",
}
DOWNPOSTFIX = '.udown'
CACHEPOSTFIX = '.ucache'
DOWN_CHUNK_SIZE = 20480
OUTFILE_WRITE_BUF = 20480
DEFAULT_CONN_TIMEOUT = 30
MODEL_DOWNLOAD_URL = "https://github.com/JanPalasek/ulozto-captcha-breaker/releases/download/v2.2/model.tflite"
TOR_DATA_DIR_PREFIX = "tor_data_dir_"
MODEL_FILENAME = "model.tflite"



from io import FileIO
from math import ceil
from typing import List
from . import const
import os
from sys import byteorder


class SegFile:
    """Implementation segment file"""
    file: str
    stat_file: str
    parts: int
    id: int

    size: int
    written: int
    pfrom: int
    pto: int

    sfp: FileIO
    sbs: int

    def __init__(self, file: str, stat_file: str, parts: int, seg_idx: int):
        self.file = file
        self.stat_file = stat_file
        self.parts = parts
        self.id = seg_idx
        self.open()

    def open(self):
        self.fp = open(self.file, 'rb+', const.OUTFILE_WRITE_BUF)
        self._load_stat()

    def _load_stat(self):
        # open stat file - must exists - buffering 0 - no need flush()
        self.sfp = open(self.stat_file, 'rb+', 0)
        # byte size of stat segment
        self.sbs = int.from_bytes(self.sfp.read(1), byteorder)
        # total_size
        total_size = int.from_bytes(self.sfp.read(self.sbs), byteorder)

        # size, from, to
        self.size = ceil(total_size / self.parts)
        self.pfrom = self.size * self.id
        self.pto = min(((self.size * (self.id + 1)) - 1), total_size)

        # fix last part have different size
        if self.id == (self.parts - 1):
            self.size = total_size - self.pfrom

        # get written
        self._read_stat()

        # seek file position
        self.fp.seek(self.cur_pos, os.SEEK_SET)

    def _read_stat(self):
        self.stat_pos = 1 + self.sbs + (self.id * self.sbs)
        self.sfp.seek(self.stat_pos, os.SEEK_SET)
        self.cur_pos = int.from_bytes(self.sfp.read(self.sbs), byteorder)
        self.written = self.cur_pos - self.pfrom

    def close(self):
        if not self.sfp.closed:
            self.sfp.close()
        if not self.fp.closed:
            self.fp.close()


class SegFileWriter(SegFile):
    """Implementation segment write file"""

    def _write_stat(self, newpos):
        self.sfp.seek(self.stat_pos, os.SEEK_SET)
        self.sfp.write(newpos.to_bytes(self.sbs, byteorder))

    def write(self, chunk):
        self.fp.seek(self.cur_pos, os.SEEK_SET)
        wrt = self.fp.write(chunk)
        self.fp.flush()
        self.written += wrt
        self.cur_pos += wrt
        self._write_stat(self.cur_pos)


class SegFileLoader:
    def __init__(self, file: str, stat_file: str, size: int, parts: int):
        self.file = file
        self.stat_file = stat_file
        self.size = size
        self.parts = parts
        self._first_created = False
        # create stat file if not exists
        self._create_files_if_not_ex()

    def make_writers(self) -> List[SegFileWriter]:
        if self._first_created:
            self.fp.close()
            self.sfp.close()
            parts = self.parts
        else:
            parts = self._get_parts_from_existing()
            self.parts = parts

        return [SegFileWriter(self.file, self.stat_file, parts, i) for i in range(parts)]

    def _get_parts_from_existing(self):
        self.sfp = open(self.stat_file, 'rb')
        sbs = int.from_bytes(self.sfp.read(1), byteorder)
        size = int.from_bytes(self.sfp.read(sbs), byteorder)

        parts_bytes_remain = self.sfp.read()  # read bytes remain
        parts = int(len(parts_bytes_remain) / sbs)
        self.part_size = ceil(size / parts)
        self.sfp.close()

        # if file size not match is not same file - truncate to fresh new
        if self.size != size:
            self._create_files_if_not_ex()
            return self.parts
        else:
            return parts

    def _create_files_if_not_ex(self):
        if not os.path.isfile(self.stat_file):
            self.fp = open(self.file, 'wb+')
            self.fp.truncate(self.size)
            self.sfp = open(self.stat_file, 'wb+')

            self._make_stat_file_data(self.size, self.parts)
            self._first_created = True

    def _make_stat_file_data(self, size, parts):
        self.part_size = ceil(size / parts)
        sb_size = ceil(size.bit_length() / 8) + 1

        self.sfp.write(sb_size.to_bytes(1, byteorder))
        self.sfp.write(size.to_bytes(sb_size, byteorder))

        for i in range(parts):
            stat_pos = 1 + sb_size + (i * sb_size)
            self.sfp.seek(stat_pos, os.SEEK_SET)
            seg = (i * self.part_size).to_bytes(sb_size, byteorder)
            self.sfp.write(seg)



import os
from queue import Queue
import requests
import threading
import time
from typing import List, Type

from uldlib.captcha import CaptchaSolver
from uldlib.const import DOWNPOSTFIX, DOWN_CHUNK_SIZE, DEFAULT_CONN_TIMEOUT
from uldlib.frontend import DownloadInfo, Frontend
from uldlib.page import Page
from uldlib.part import DownloadPart
from uldlib.segfile import SegFileLoader
from uldlib.torrunner import TorRunner
from uldlib.utils import DownloaderError, DownloaderStopped, LogLevel


class Downloader:
    terminating: bool

    threads: List[threading.Thread]
    stop_download: threading.Event

    frontend: Type[Frontend]
    frontend_thread: threading.Thread = None
    stop_frontend: threading.Event

    captcha_solver: Type[CaptchaSolver]
    captcha_thread: threading.Thread = None
    stop_captcha: threading.Event

    download_url_queue: Queue
    parts: int
    tor: TorRunner
    page: Page

    password: str

    def __init__(self, tor: TorRunner, frontend: Type[Frontend], captcha_solver: Type[CaptchaSolver]):
        """Initialize the Downloader.

           The TorRunner could be launched or not, the .launch() method will be called when needed.
           Also it is caller responsibility to call .stop() method on the TorRunner.
        """

        self.success = None
        self.url = None
        self.filename = None
        self.output_filename = None
        self.stat_filename = None
        self.total_size = None
        self.frontend = frontend
        self.log = frontend.main_log
        self.captcha_solver = captcha_solver

        self.cli_initialized = False
        self.conn_timeout = None
        self.tor = tor

    def terminate(self, quiet: bool = False):
        if self.terminating:
            return
        self.terminating = True

        if not quiet:
            self.log('Terminating download. Please wait for stopping all threads.', level=LogLevel.WARNING)

        self.stop_download.set()
        self.stop_captcha.set()
        if self.captcha_thread and self.captcha_thread.is_alive():
            self.captcha_thread.join()
        for p in self.threads:
            if p.is_alive():
                p.join()

        if not quiet:
            self.log('Download terminated.', level=LogLevel.WARNING)

        self.stop_frontend.set()
        if self.frontend_thread and self.frontend_thread.is_alive():
            self.frontend_thread.join()

    def clean(self):
        # remove resume .udown file
        if os.path.exists(self.stat_filename):
            os.remove(self.stat_filename)
        if self.page.linkCache is not None:
            self.page.linkCache.delete_cache_file()

    def _captcha_breaker(self, page, parts):
        msg = ""
        if page.isDirectDownload:
            msg = "Solve direct dlink .."
        else:
            msg = "Solve CAPTCHA dlink .."

        try:
            for url in self.captcha_download_links_generator:
                if self.stop_captcha.is_set():
                    break
                self.captcha_solver.log(msg)
                self.download_url_queue.put(url)
        except DownloaderError as e:
            self.captcha_solver.log(str(e), level=LogLevel.ERROR)

    def _download_part(self, part: DownloadPart):
        try:
            self._download_part_internal(part)
        except Exception as e:
            part.exception = e
            part.set_status(f"Error: {e}", error=True)

    def _download_part_internal(self, part: DownloadPart):
        """Download given part of the download.

            Arguments:
                part (DownloadPart): Specification of the part to download
        """

        writer = part.writer

        part.lock.acquire()
        part.started = True
        part.start_time = time.time()
        part.lock.release()

        while True:
            if self.stop_download.is_set():
                return

            part.set_status("Starting download")
            # Note the stream=True parameter
            r = requests.get(part.download_url, stream=True, allow_redirects=True, headers={
                "Range": "bytes={}-{}".format(writer.pfrom + writer.written, writer.pto),
                "Connection": "close",
            })

            if r.status_code == 429:
                part.set_status("Status code 429 Too Many Requests returned… will try again in few seconds", warning=True)
                time.sleep(5)
            elif r.status_code == 425:
                part.set_status("Status code 425 Too Early returned… will try again in few seconds", warning=True)
                time.sleep(5)
            else:
                break

        if r.status_code != 206 and r.status_code != 200:
            part.set_status(f"Status code {r.status_code} returned: {writer.pfrom + writer.written}/{writer.pto}", error=True)
            return

        part.set_status("")

        # reimplement as multisegment write file class
        for chunk in r.iter_content(chunk_size=DOWN_CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                writer.write(chunk)

                part.lock.acquire()
                part.d_now += len(chunk)
                part.d_total += len(chunk)
                part.lock.release()

                if self.stop_download.is_set():
                    # TODO: cancel the request when urllib3/requests will be able to do so
                    # (now r.close() would be blocking until all chunks downloaded)
                    return

        # download end status
        r.close()
        part.lock.acquire()
        part.completed = True
        part.completion_time = time.time()
        part.lock.release()

        # close part file files
        writer.close()

        # reuse download link if need
        self.download_url_queue.put(part.download_url)

    def download(self, url: str, parts: int = 10, password: str = "", target_dir: str = "", temp_dir: str = "", do_overwrite: bool = False, conn_timeout=DEFAULT_CONN_TIMEOUT):
        """Download file from Uloz.to using multiple parallel downloads.
            Arguments:
                url: URL of the Uloz.to file to download
                parts: Number of parts that will be downloaded in parallel (default: 10)
                target_dir: Directory where the download should be saved (default: current directory)
                do_overwrite: Overwrite files without asking
                temp_dir: Directory where temporary files will be created (default: current directory)
                password: Optional password to access the Uloz.to file
        """
        self.url = url
        self.parts = parts
        self.conn_timeout = conn_timeout

        self.threads = []
        self.terminating = False
        self.isLimited = False
        self.isCaptcha = False

        self.stop_download = threading.Event()
        self.stop_captcha = threading.Event()
        self.stop_frontend = threading.Event()

        # 1. Prepare downloads
        self.log("Starting downloading for url '{}'".format(url))
        # 1.1 Get all needed information
        self.log("Getting info (filename, filesize, …)")

        try:
            self.page = Page(
                url=url,
                temp_dir=temp_dir,
                parts=parts,
                password=password,
                frontend=self.frontend,
                tor=self.tor,
                conn_timeout=self.conn_timeout,
            )
            page = self.page  # shortcut
            page.parse()

        except Exception as e:
            raise DownloaderError('Cannot download file: ' + str(e))

        # Check of the target is a file or directory and construct the output path accordingly
        if not os.path.isdir(target_dir) and target_dir[-1] != '/':
            # Path to a file has been provided
            self.output_filename = target_dir
        else:
            # Create the output directory if does not exist
            os.makedirs(target_dir, exist_ok = True)
            self.output_filename = os.path.join(target_dir, page.filename)
        self.filename = os.path.basename(self.output_filename)
        self.stat_filename = os.path.join(temp_dir, self.filename + DOWNPOSTFIX)

        self.log("Downloading into: '{}'".format(self.output_filename))

        # Do check - only if .udown status file not exists get question
        # .udown file is always present in cli_mode = False
        if os.path.isfile(self.output_filename) and not os.path.isfile(self.stat_filename):
            if self.frontend.supports_prompt and not do_overwrite:
                answer = self.frontend.prompt(
                    "WARNING: File '{}' already exists, overwrite it? [y/n] ".format(self.output_filename), level=LogLevel.WARNING)
                if answer != 'y':
                    raise DownloaderStopped()
            else:
                self.log("WARNING: File '{}' already exists, but .udown file not present. File will be overwritten.."
                         .format(self.output_filename), level=LogLevel.WARNING)

        info = DownloadInfo()
        info.filename = self.filename
        info.url = page.url

        if page.quickDownloadURL is not None:
            self.log("You are VERY lucky, this is QUICK direct download without CAPTCHA, downloading as 1 quick part :)")
            info.download_type = "fullspeed direct download (without CAPTCHA)"
            download_url = page.quickDownloadURL
            self.captcha_solve_func = None

        if page.slowDownloadURL is not None:
            self.isLimited = True
            if page.isDirectDownload:
                self.log("You are lucky, this is slow direct download without CAPTCHA :)")
                info.download_type = "slow direct download (without CAPTCHA)"
            else:
                self.isCaptcha = True
                self.log("CAPTCHA protected download - CAPTCHA challenges will be displayed")
                info.download_type = "CAPTCHA protected download"

            if self.isCaptcha and self.captcha_solver.cannot_solve:
                raise DownloaderError("Cannot solve CAPTCHAs, no solver available. Terminating")

            self.captcha_download_links_generator = page.captcha_download_links_generator(
                solver=self.captcha_solver, stop_event=self.stop_captcha,
            )
            download_url = next(self.captcha_download_links_generator)

        head = requests.head(download_url, allow_redirects=True)
        self.total_size = int(head.headers['Content-Length'])

        try:
            file_data = SegFileLoader(self.output_filename, self.stat_filename, self.total_size, parts)
            writers = file_data.make_writers()
        except Exception as e:
            raise DownloaderError(f"Failed: Can not create '{self.output_filename}' error: {e} ")

        info.total_size = self.total_size
        info.part_size = file_data.part_size
        info.parts = file_data.parts

        downloads: List[DownloadPart] = [DownloadPart(w) for w in writers]

        # 2. All info gathered, initialize frontend

        self.log("Download in progress")
        # fill placeholder before download started
        for part in downloads:
            if page.isDirectDownload:
                part.set_status("Waiting for direct link…")
            else:
                part.set_status("Waiting for CAPTCHA…")

        self.frontend_thread = threading.Thread(
            target=self.frontend.run,
            args=(info, downloads, self.stop_frontend, self.terminate)
        )
        self.frontend_thread.start()

        # Prepare queue for recycling download URLs
        self.download_url_queue = Queue(maxsize=0)

        # limited must use TOR and solve links or captcha
        if self.isLimited:
            # Reuse already solved links
            self.download_url_queue.put(download_url)

            # Start CAPTCHA breaker in separate process
            self.captcha_thread = threading.Thread(
                target=self._captcha_breaker, args=(page, self.parts)
            )

        cpb_started = False
        page.alreadyDownloaded = 0

        # 3. Start all downloads fill self.threads
        for part in downloads:
            if self.terminating:
                return

            if part.writer.written == part.writer.size:
                part.completed = True
                part.set_status("Already downloaded from previous run, skipping")
                page.alreadyDownloaded += 1
                continue

            if self.isLimited:
                if not cpb_started:
                    self.captcha_thread.start()
                    cpb_started = True
                part.download_url = self.download_url_queue.get()
            else:
                part.download_url = download_url

            # Start download process in another process (parallel):
            t = threading.Thread(target=self._download_part, args=(part,))
            t.start()
            self.threads.append(t)

        if self.isLimited:
            # no need for another CAPTCHAs
            self.stop_captcha.set()
            if self.isCaptcha:
                self.captcha_solver.log("All downloads started, no need to solve another CAPTCHAs…")
            else:
                self.captcha_solver.log("All downloads started, no need to solve another direct links…")

        # 4. Wait for all downloads to finish
        success = True
        for (t, part) in zip(self.threads, downloads):
            while t.is_alive():
                t.join(1)
            if part.error:
                success = False

        self.terminate(quiet=True)

        self.success = success
        # result end status
        if not success:
            raise DownloaderError("Failure of one or more downloads, exiting")

        self.log("All downloads successfully finished", level=LogLevel.SUCCESS)



from abc import abstractmethod
from datetime import datetime, timedelta
from io import TextIOWrapper
from traceback import print_exc
import colors
import os
import sys
import time
import threading
import json
from typing import Dict, List, Optional, Tuple

from uldlib import utils
from uldlib.const import CLI_STATUS_STARTLINE
from uldlib.part import DownloadPart
from uldlib.utils import LogLevel, Status


class DownloadInfo:
    filename: str
    url: str
    download_type: str
    total_size: int
    part_size: int
    parts: int


class Frontend():
    supports_prompt: bool

    def __init__(self, supports_prompt: bool):
        self.supports_prompt = supports_prompt

    # Log functions for TOR, CAPTCHA solver and for everything else
    # - `progress` is used by some frontends, when True it means that this message
    #   is progress (progress bar or similar) and should not be logged / should be overwritten

    @abstractmethod
    def tor_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        pass

    @abstractmethod
    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        pass

    @abstractmethod
    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        pass

    @abstractmethod
    def captcha_stats(self, stats: Dict[str, int]):
        pass

    @abstractmethod
    def prompt(self, msg: str, level: LogLevel = LogLevel.INFO) -> str:
        pass

    @abstractmethod
    def run(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event, terminate_func):
        pass


class ConsoleFrontend(Frontend):
    cli_initialized: bool
    show_parts: bool
    logfile: Optional[TextIOWrapper] = None

    last_log: Tuple[str, LogLevel]

    last_captcha_log: Tuple[str, LogLevel]
    last_captcha_stats: Dict[str, int]

    def __init__(self, show_parts: bool = False, logfile: str = ""):
        super().__init__(supports_prompt=True)
        self.cli_initialized = False
        self.last_log = ("", LogLevel.INFO)
        self.last_captcha_log = ("", LogLevel.INFO)
        self.last_captcha_stats = None
        self.show_parts = show_parts
        if logfile:
            self.logfile = open(logfile, 'a')

    def __del__(self):
        if self.logfile:
            self.logfile.close()

    @staticmethod
    def _log_print(msg: str, progress: bool):
        if progress:
            sys.stdout.write(msg + "\033[K\r")
        else:
            print(msg)

    def _log_logfile(self, prefix: str, msg: str, progress: bool, level: LogLevel):
        if progress or self.logfile is None:
            return

        t = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.logfile.write(f"{t} {prefix}\t[{level.name}] {msg}\n")
        self.logfile.flush()

    def tor_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        self.last_captcha_log = (msg, level)  # shares same log with CAPTCHA
        self._log_logfile('TOR ', msg, progress=progress, level=level)
        if not self.cli_initialized:
            self._log_print(colors.blue("[TOR]\t") + utils.color(msg, level), progress=progress)

    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        self.last_captcha_log = (msg, level)
        self._log_logfile('CAPTCHA', msg, progress=progress, level=level)
        if not self.cli_initialized:
            self._log_print(colors.blue("[Link solve]\t") + utils.color(msg, level), progress=progress)

    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        self.last_log = (msg, level)
        self._log_logfile('MAIN', msg, progress=progress, level=level)
        if not self.cli_initialized:
            self._log_print(utils.color(msg, level), progress=progress)

    def captcha_stats(self, stats: Dict[str, int]):
        self.last_captcha_stats = stats

    def prompt(self, msg: str, level: LogLevel = LogLevel.INFO) -> str:
        print(utils.color(msg, level), end="")
        return input().strip()

    @staticmethod
    def _stat_fmt(stats: Dict[str, int]):
        count = colors.blue(stats['all'])
        ok = colors.green(stats['ok'])
        bad = colors.red(stats['bad'])
        lim = colors.red(stats['lim'])
        blo = colors.red(stats['block'])
        net = colors.red(stats['net'])
        return f"[Ok: {ok} / {count}] :( [Badcp: {bad} Limited: {lim} Censored: {blo} NetErr: {net}]"

    @staticmethod
    def _print(text, x=0, y=0):
        sys.stdout.write("\033[{};{}H".format(y, x))
        sys.stdout.write("\033[K")
        sys.stdout.write(text)
        sys.stdout.flush()

    def run(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event, terminate_func):
        try:
            self._loop(info, parts, stop_event)
        except Exception:
            if self.cli_initialized:
                y = info.parts + CLI_STATUS_STARTLINE + 4
                sys.stdout.write("\033[{};{}H".format(y, 0))
                sys.stdout.write("\033[?25h")  # show cursor
                self.cli_initialized = False
                print("")
            print_exc()
            terminate_func()

    def _loop(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event):
        os.system('cls' if os.name == 'nt' else 'clear')
        sys.stdout.write("\033[?25l")  # hide cursor
        self.cli_initialized = True

        print(colors.blue("File:\t\t") + colors.bold(info.filename))
        print(colors.blue("URL:\t\t") + info.url)
        print(colors.blue("Download type:\t") + info.download_type)
        print(colors.blue("Size / parts: \t") +
              colors.bold(f"{round(info.total_size / 1024**2, 2)}MB => " +
              f"{info.parts} x {round(info.part_size / 1024**2, 2)}MB"))

        t_start = time.time()
        s_start = 0
        for part in parts:
            (_, _, size) = part.get_frontend_status()
            s_start += size
        last_bps = [(s_start, t_start)]

        y = 0

        while True:
            t = time.time()
            # Get parts info
            lines = []
            s = 0
            for part in parts:
                (line, level, size) = part.get_frontend_status()
                lines.append(utils.color(line, level))
                s += size

            y = CLI_STATUS_STARTLINE

            # Print CAPTCHA/TOR status
            (msg, level) = self.last_captcha_log
            self._print(
                colors.yellow("[Link solve]\t") +
                utils.color(msg, level),
                y=y
            )
            y += 1
            if self.last_captcha_stats is not None:
                self._print(
                    colors.yellow("\t\t") + self._stat_fmt(self.last_captcha_stats),
                    y=y
                )
                y += 1

            # Print overall progress line
            if t == t_start:
                total_bps = 0
                now_bps = 0
            else:
                total_bps = (s - s_start) / (t - t_start)
                # Average now bps for last 10 measurements
                if len(last_bps) >= 10:
                    last_bps = last_bps[1:]
                (s_last, t_last) = last_bps[0]
                now_bps = (s - s_last) / (t - t_last)
                last_bps.append((s, t))

            remaining = (info.total_size - s) / total_bps if total_bps > 0 else 0

            self._print(colors.yellow(
                f"[Progress]\t"
                f"{(s / 1024 ** 2):.2f} MB"
                f" ({(s / info.total_size * 100):.2f} %)"
                f"\tavg. speed: {(total_bps / 1024 ** 2):.2f} MB/s"
                f"\tcurr. speed: {(now_bps / 1024 ** 2):.2f} MB/s"
                f"\tremaining: {timedelta(seconds=round(remaining))}"),
                y=y
            )
            y += 1

            # Print last log message
            (msg, level) = self.last_log
            self._print(
                colors.yellow("[STATUS]\t") +
                (msg if level == LogLevel.INFO else colors.negative(utils.color(msg, level))),
                y=y
            )
            y += 1

            # Print parts
            if self.show_parts:
                for (line, part) in zip(lines, parts):
                    self._print(
                        colors.blue(f"[Part {part.id}]") + f"\t{line}",
                        y=(y + part.id))

            if stop_event.is_set():
                break

            time.sleep(0.5)

        if self.cli_initialized:
            y = info.parts + CLI_STATUS_STARTLINE + 4
            sys.stdout.write("\033[{};{}H".format(y + 2, 0))
            sys.stdout.write("\033[?25h")  # show cursor
            self.cli_initialized = False

        elapsed = time.time() - t_start
        # speed in bytes per second:
        speed = (s - s_start) / elapsed if elapsed > 0 else 0
        print(colors.blue("Statistics:\t") + "Downloaded {}{} MB in {} (average speed {} MB/s)".format(
            round((s - s_start) / 1024**2, 2),
            "" if s_start == 0 else (
                "/"+str(round(info.total_size / 1024**2, 2))
            ),
            str(timedelta(seconds=round(elapsed))),
            round(speed / 1024**2, 2)
        ))


class JSONFrontend(Frontend):
    show_parts: bool
    logfile: Optional[TextIOWrapper] = None

    last_log: Tuple[str, LogLevel]

    last_captcha_log: Tuple[str, LogLevel]
    last_captcha_stats: Dict[str, int]

    def __init__(self, show_parts: bool = False, logfile: str = ""):
        super().__init__(supports_prompt=True)
        self.last_log = ("", LogLevel.INFO)
        self.last_captcha_log = ("", LogLevel.INFO)
        self.last_captcha_stats = None
        self.show_parts = show_parts
        if logfile:
            self.logfile = open(logfile, 'a')
        print(json.dumps({"status": Status.INITIALIZING}))

    def __del__(self):
        if self.logfile:
            self.logfile.close()

    @staticmethod
    def _log_print(msg: str, progress: bool):
        if progress:
            sys.stdout.write(msg + "\033[K\r")
        else:
            print(msg)

    def _log_logfile(self, prefix: str, msg: str, progress: bool, level: LogLevel):
        if progress or self.logfile is None:
            return

        t = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.logfile.write(f"{t} {prefix}\t[{level.name}] {msg}\n")
        self.logfile.flush()

    def tor_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        print(json.dumps({"tor": msg}))
        pass

    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        print(json.dumps({"captcha": msg}))
        pass

    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
        if LogLevel == LogLevel.ERROR:
            print(json.dumps({"status": Status.ERROR, "message": msg}))

    def captcha_stats(self, stats: Dict[str, int]):
        self.last_captcha_stats = stats

    def prompt(self, msg: str, level: LogLevel = LogLevel.INFO) -> str:
        # print(utils.color(msg, level), end="")
        return input().strip()

    def run(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event, terminate_func):
        try:
            self._loop(info, parts, stop_event)
        except Exception:
            print_exc()
            terminate_func()

    def _loop(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event):
        jsonReport = JSONReport(info)

        t_start = time.time()
        s_start = 0
        for part in parts:
            (_, _, size) = part.get_frontend_status()
            s_start += size
        last_bps = [(s_start, t_start)]

        while True:
            t = time.time()
            # Get parts info
            s = 0
            for part in parts:
                (line, level, size) = part.get_frontend_status()
                s += size

            # Print overall progress line
            if t == t_start:
                total_bps = 0
                now_bps = 0
            else:
                total_bps = (s - s_start) / (t - t_start)
                # Average now bps for last 10 measurements
                if len(last_bps) >= 10:
                    last_bps = last_bps[1:]
                (s_last, t_last) = last_bps[0]
                now_bps = (s - s_last) / (t - t_last)
                last_bps.append((s, t))

            jsonReport.update(s, total_bps, now_bps)
            print(jsonReport)

            if stop_event.is_set():
                break

            time.sleep(0.5)

        elapsed = time.time() - t_start
        # speed in bytes per second:
        speed = (s - s_start) / elapsed if elapsed > 0 else 0

        print(json.dumps({"status": Status.COMPLETED, "duration": str(timedelta(seconds=round(elapsed))), "avg_speed": f"{round(speed / 1024**2, 2)} MB/s"}))


class JSONReport:
    status: str
    file: str
    url: str
    size: str
    downloaded: str
    percent: str
    avg_speed: str
    curr_speed: str
    remaining: str
    size_float: float

    def __init__(self, info: DownloadInfo) -> None:
        self.status = Status.DOWNLOADING
        self.file = info.filename,
        self.url = info.url,
        self.size = f"{round(info.total_size / 1024**2, 2)} MB"
        self.size_float = info.total_size

    def update(self, down_size, total_bps, now_bps):
        self.downloaded = f"{(down_size / 1024 ** 2):.2f} MB",
        self.percent = f"{(down_size / self.size_float * 100):.2f} %",
        self.avg_speed = f"{(total_bps / 1024 ** 2):.2f} MB/s",
        self.curr_speed = f"{(now_bps / 1024 ** 2):.2f} MB/s",

        remaining = (self.size_float - down_size) / total_bps if total_bps > 0 else 0
        self.remaining = f"{timedelta(seconds=round(remaining))}"

    def __str__(self) -> str:
        return json.dumps(self.__dict__).replace('["', '"').replace('"]', '"')



from abc import abstractmethod
import threading
import time
import importlib
from typing import Dict
import requests
from PIL import Image
from io import BytesIO

from uldlib.frontend import Frontend
from uldlib.utils import LogLevel


class CaptchaSolver():
    frontend: Frontend
    cannot_solve: bool = False

    def __init__(self, frontend: Frontend):
        self.frontend = frontend

    def log(self, msg: str, level: LogLevel = LogLevel.INFO):
        self.frontend.captcha_log(msg, level)

    def stats(self, stats: Dict[str, int]):
        self.frontend.captcha_stats(stats)

    @abstractmethod
    def solve(self, img_url: str, stop_event: threading.Event = None) -> str:
        pass


class Dummy(CaptchaSolver):
    """Dumy solver when tflite_runtime nor tkinter available."""

    def __init__(self, frontend):
        super().__init__(frontend)
        self.cannot_solve = True

    def solve(self, img_url, stop_event):
        pass


class ManualInput(CaptchaSolver):
    """Display captcha from given URL and ask user for input in GUI window."""

    def __init__(self, frontend):
        super().__init__(frontend)

    def solve(self, img_url: str, stop_event: threading.Event = None) -> str:
        import tkinter as tk
        from PIL import ImageTk

        root = tk.Tk()
        root.focus_force()
        root.title("Opiš kód z obrázku")
        # use width x height + x_offset + y_offset (no spaces!)
        root.geometry("300x140")

        # make window popup on top
        root.lift()
        root.attributes('-topmost', True)
        root.after(1, lambda: root.focus_force())

        def disable_event():
            pass

        root.protocol("WM_DELETE_WINDOW", disable_event)

        u = requests.get(img_url)
        raw_data = u.content

        im = Image.open(BytesIO(raw_data))
        photo = ImageTk.PhotoImage(im)
        label = tk.Label(image=photo)
        label.image = photo
        label.pack()

        entry = tk.Entry(root)
        entry.pack()
        entry.bind('<Return>', lambda event: root.quit())
        entry.after(1, lambda: entry.focus_set())

        tk.Button(root, text='Send', command=root.quit).pack()

        # Closing of the window separated to thread because it can be closed by
        # the user input (done==True) or by the terminating application (stop_event)
        done = False

        def stop_func():
            while True:
                if done or (stop_event and stop_event.is_set()):
                    break
                time.sleep(0.1)
            self.log("Closing tkinter window, wait…")
            root.quit()

        stop_thread = threading.Thread(target=stop_func)
        stop_thread.start()
        root.mainloop()  # Wait for user input

        value = entry.get()
        done = True
        stop_thread.join()
        root.destroy()
        return value


class AutoReadCaptcha(CaptchaSolver):
    def __init__(self, model_path: str, model_url: str, frontend: Frontend):
        super().__init__(frontend)

        from urllib.request import urlretrieve
        import os

        tfull_available = importlib.util.find_spec('tensorflow') and importlib.util.find_spec('tensorflow.lite')
        tflite_available = importlib.util.find_spec('tflite_runtime')

        if tfull_available:
            import tensorflow.lite as tflite
        elif tflite_available:
            import tflite_runtime.interpreter as tflite
        else:
            raise ImportError('No tensorflow.lite or tflite_runtime available.')

        def reporthook(blocknum, block_size, total_size):
            """
            Credits to jfs from https://stackoverflow.com/questions/13881092/download-progressbar-for-python-3
            """
            readsofar = blocknum * block_size
            if total_size > 0:
                percent = readsofar * 1e2 / total_size
                self.log("Downloading model from %s: %5.1f%% %*d / %d" % (
                    model_url, percent, len(str(total_size)), readsofar, total_size))
            else:  # total size is unknown
                self.log("Downloading model from %s: read %d" % (model_url, readsofar))

        if not os.path.exists(model_path):
            self.log(f"Downloading model from {model_url}")
            # download into temp model in order to detect incomplete downloads
            model_temp_path = f"{model_path}.tmp"
            urlretrieve(model_url, model_temp_path, reporthook)
            self.log("Downloading of the model finished")

            # rename temp model
            os.rename(model_temp_path, model_path)

        model_content = open(model_path, "rb").read()
        self.interpreter = tflite.Interpreter(model_content=model_content)

    def solve(self, img_url, stop_event=None) -> str:
        # stop_event not used, because tflite interpreter is hard to cancel (but is is quick)
        import numpy as np

        interpreter = self.interpreter

        self.log("Auto solving CAPTCHA")

        u = requests.get(img_url)
        raw_data = u.content

        img = Image.open(BytesIO(raw_data))
        img = np.asarray(img)

        # normalize to [0...1]
        img = (img / 255).astype(np.float32)

        # convert to grayscale
        r, g, b = img[:, :, 0], img[:, :, 1], img[:, :, 2]
        input = 0.299 * r + 0.587 * g + 0.114 * b

        # input has nowof  shape (70, 175)
        # we modify dimensions to match model's input
        input = np.expand_dims(input, 0)
        input = np.expand_dims(input, -1)
        # input is now of shape (batch_size, 70, 175, 1)
        # output will have shape (batch_size, 4, 26)

        interpreter.allocate_tensors()
        input_details = interpreter.get_input_details()
        output_details = interpreter.get_output_details()
        interpreter.set_tensor(input_details[0]['index'], input)
        interpreter.invoke()

        # predict and get the output
        output = interpreter.get_tensor(output_details[0]['index'])
        # now get labels
        labels_indices = np.argmax(output, axis=2)

        available_chars = "abcdefghijklmnopqrstuvwxyz"

        def decode(li):
            result = []
            for char in li:
                result.append(available_chars[char])
            return "".join(result)

        decoded_label = [decode(x) for x in labels_indices][0]
        self.log(f"CAPTCHA auto solved as '{decoded_label}'")
        return decoded_label



import socket
from enum import Enum
from typing import List

from colors import colors


class LogLevel(Enum):
    INFO = 1
    WARNING = 2
    ERROR = 3
    SUCCESS = 4


class Status(str, Enum):
    INITIALIZING: str = "initializing"
    DOWNLOADING: str = "downloading"
    COMPLETED: str = "completed"
    ERROR: str = "error"


def color(text: str, level: LogLevel) -> str:
    if level == LogLevel.WARNING:
        return colors.yellow(text)
    if level == LogLevel.ERROR:
        return colors.red(text)
    if level == LogLevel.SUCCESS:
        return colors.green(text)
    return text


def _is_port_available(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('127.0.0.1', port)) != 0


def get_available_port(given_port: int, skip: List[int] = []) -> int:
    max_attempts = 65535
    while given_port < max_attempts:
        if _is_port_available(given_port) and given_port not in skip:
            return given_port
        given_port += 1
    else:
        raise ValueError("Cannot get available port")


class DownloaderStopped(Exception):
    pass


class DownloaderError(Exception):
    pass



import re
import threading
from typing import Optional, Type
from urllib.parse import urlparse, urljoin
from os import path
import requests

from uldlib.captcha import CaptchaSolver
from uldlib.frontend import LogLevel, Frontend
from uldlib.torrunner import TorRunner

from .const import XML_HEADERS, DEFAULT_CONN_TIMEOUT
from .linkcache import LinkCache

from requests.sessions import RequestsCookieJar


def parse_single(text, regex):
    p = re.compile(regex, re.IGNORECASE)
    result = p.findall(text)
    if len(result) == 0:
        return None
    return result[0]


def strip_tracking_info(url: str):
    return url.split("#!")[0] if "#!" in url else url


class Page:
    url: str
    body: str
    baseURL: str
    slug: str
    pagename: str

    filename: str
    slowDownloadURL: str
    quickDownloadURL: str
    captchaURL: str
    isDirectDownload: bool
    numTorLinks: int
    alreadyDownloaded: int
    password: str

    needPassword: bool = False

    linkCache: Optional[LinkCache] = None

    def __init__(self, url: str, temp_dir: str, parts: int, password: str, frontend: Frontend, tor: TorRunner, conn_timeout=DEFAULT_CONN_TIMEOUT):
        """Check given url and if it looks ok GET the Uloz.to page and save it.

            Arguments:
                url (str): URL of the page with file
                temp_dir (str): directory where .ucache file will be created
                parts (int): number of segments (parts)
                password (str): password to access the Uloz.to file
                frontend (Frontend): frontend object for password prompt (if supported)
                tor (TorRunner): tor runner instance
            Raises:
                RuntimeError: On invalid URL, deleted file or other error related to getting the page.
        """

        self.url = strip_tracking_info(url)
        self.temp_dir = temp_dir
        self.parts = parts
        self.password = password
        self.frontend = frontend
        self.tor = tor
        self.conn_timeout = conn_timeout

        self.tor.launch()  # ensure that TOR is running

        parsed_url = urlparse(self.url)
        if parsed_url.scheme == "":
            # Common mistake when copied from addres bar without https://
            self.url = "https://" + self.url
            parsed_url = urlparse(self.url)

        if not parsed_url.hostname:
            raise RuntimeError(f"Invalid url {self.url}")

        self.pagename = parsed_url.hostname.capitalize()
        self.cli_initialized = False
        self.alreadyDownloaded = 0
        self.stats = {"all": 0, "ok": 0, "bad": 0,
                      "lim": 0, "block": 0, "net": 0}  # statistics

        s = requests.Session()
        s.proxies = self.tor.proxies
        # special case for Pornfile.cz run by Uloz.to - confirmation is needed
        if parsed_url.hostname == "pornfile.cz":
            s.post("https://pornfile.cz/porn-disclaimer/", data={
                "agree": "Souhlasím",
                "_do": "pornDisclaimer-submit",
            })

        # If file is file-tracking link we need to get normal file link from it
        if url.startswith('{uri.scheme}://{uri.netloc}/file-tracking/'.format(uri=parsed_url)):
            s.get(url, allow_redirects=False)
            if 'Location' in r.headers:
                self.url = strip_tracking_info(r.headers['Location'])
                parsed_url = urlparse(self.url)

        r = s.get(self.url)
        self.baseURL = "{uri.scheme}://{uri.netloc}".format(uri=parsed_url)

        if r.status_code == 451:
            raise RuntimeError(
                f"File was deleted from {self.pagename} due to legal reasons (status code 451)")
        elif r.status_code == 401:
            self.needPassword = True
            r = self.enter_password(s)
        elif r.status_code == 404:
            raise RuntimeError(
                f"{self.pagename} returned status code {r.status_code}, file does not exist")
        elif r.status_code != 200:
            raise RuntimeError(
                f"{self.pagename} returned status code {r.status_code}, error: {r.text}")

        # Get file slug from URL
        self.slug = parse_single(parsed_url.path, r'/file/([^\\]*)/')
        if self.slug is None:
            raise RuntimeError(
                f"Cannot parse file slug from {self.pagename} URL")

        self.body = r.text

    def parse(self):
        """Try to parse all information from the page (filename, download links, ...)

        Raises:
            RuntimeError: When mandatory fields cannot be parsed.
        """

        # Parse filename only to the first | (Uloz.to sometimes add titles like "name | on-line video | Ulož.to" and so on)
        self.filename = parse_single(self.body, r'<title>([^\|]*)\s+\|.*</title>')

        # Replace illegal characters in filename https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
        self.filename = re.sub(r'[<>:,\"/\\|\?*]', "-", self.filename)

        download_found = False

        self.quickDownloadURL = parse_single(self.body, r'href="(/quickDownload/[^"]*)"')
        if self.quickDownloadURL:
            download_found = True
            self.quickDownloadURL = self.baseURL + self.quickDownloadURL

        # detect direct download from self.body
        isDirect = parse_single(
            self.body, r'data-href="/download-dialog/free/[^"]+" +class=".+(js-free-download-button-direct).+"')

        if isDirect == 'js-free-download-button-direct':
            self.isDirectDownload = True
        else:
            self.isDirectDownload = False

        # Other files are protected by CAPTCHA challenge
        # <a href="javascript:;" data-href="/download-dialog/free/default?fileSlug=apj0q49iETRR" class="c-button c-button__c-white js-free-download-button-dialog t-free-download-button">
        self.captchaURL = parse_single(self.body, r'data-href="(/download-dialog/free/[^"]*)"')
        if self.captchaURL:
            download_found = True
            self.captchaURL = self.baseURL + self.captchaURL
        self.slowDownloadURL = self.captchaURL

        # Check if slowDirectDownload or form data for CAPTCHA was parsed
        if not download_found:
            raise RuntimeError(f"Cannot parse {self.pagename} page to get download information,"
                               + " no direct download URL and no CAPTCHA challenge URL found")

    # print TOR network error and += stats
    def _error_net_stat(self, err, log_func):
        self.stats["all"] += 1
        log_func(f"Network error get new TOR connection: {err}", level=LogLevel.ERROR)
        self.stats["net"] += 1

    def _link_validation_stat(self, resp, log_func):
        linkdata = resp.text
        self.stats["all"] += 1
        ok = False
        reload = True

        # search errors..
        good_str = "afterDownloadUrl"
        blk_str = 'blocked'
        lim_str = 'limit-exceeded'
        bcp_str = "formErrorContent"

        # msgs
        lim_msg = "IP limited TOR exit.. get new TOR session"
        blk_msg = "Blocked TOR exit IP.. get new TOR session"
        bcp_msg = "Bad captcha.. Try again using same IP"

        if good_str in linkdata:
            self.stats["ok"] += 1
            ok = True
        elif lim_str in linkdata:
            self.stats["lim"] += 1
            if not self.isDirectDownload:
                log_func(lim_msg, level=LogLevel.ERROR)
        elif blk_str in linkdata:
            self.stats["block"] += 1
            if not self.isDirectDownload:
                log_func(blk_msg, level=LogLevel.ERROR)
        elif bcp_str in linkdata:
            self.stats["bad"] += 1
            log_func(bcp_msg, level=LogLevel.ERROR)
            reload = False  # bad captcha same IP again

        return (ok, reload)

    def captcha_download_links_generator(self, solver: Type[CaptchaSolver], stop_event: threading.Event = None):
        """
            Generator for CAPTCHA download links using Tor sessions.
            Get download link by solving CAPTCHA, calls CAPTCHA related functions..

            Arguments:
                solver (CaptchaSolver): Class with solve method which gets CAPTCHA challenge URL and returns CAPTCHA answer
                stop_event: Threading event to check when to stop

            Returns:
                str: URL for downloading the file
        """

        self.numTorLinks = 0
        self.cacheEmpty = False
        self.linkCache = LinkCache(path.join(self.temp_dir, self.filename))

        while not self.cacheEmpty:
            cached = self.linkCache.get_all_valid_links()
            self.cacheEmpty = True
            for link in cached:
                # linkCache.use(self.numTorLinks)
                self.numTorLinks += 1
                yield link

        while (self.numTorLinks + self.alreadyDownloaded) < self.parts:
            if stop_event and stop_event.is_set():
                break

            self.tor.launch()  # ensure that TOR is running

            # reload tor after 1. use or all except badCaptcha case
            reload = False
            if self.stats["all"] > 0 or reload:
                self.tor.reload()

            try:
                s = requests.Session()
                s.proxies = self.tor.proxies
                if urlparse(self.url).hostname == "pornfile.cz":
                    r = s.post("https://pornfile.cz/porn-disclaimer/", data={
                        "agree": "Souhlasím",
                        "_do": "pornDisclaimer-submit",
                    })

                if self.needPassword:
                    s.get(self.url)  # to obtain initial set of cookies
                    s.post(self.url, data={
                        "password": self.password,
                        "password_send": "Odeslat",
                        "_do": "passwordProtectedForm-submit",
                    })

                resp = requests.Response()

                if self.isDirectDownload:
                    solver.log(f"TOR get downlink (timeout {self.conn_timeout})")
                    resp = s.get(self.captchaURL,
                                 headers=XML_HEADERS, timeout=self.conn_timeout)
                else:
                    solver.log(f"TOR get new CAPTCHA (timeout {self.conn_timeout})")
                    r = s.get(self.captchaURL, headers=XML_HEADERS)

                    # <img class="xapca-image" src="//xapca1.uloz.to/0fdc77841172eb6926bf57fe2e8a723226951197/image.jpg" alt="">
                    captcha_image_url = parse_single(
                        r.text, r'<img class="xapca-image" src="([^"]*)" alt="">')
                    if captcha_image_url is None:
                        solver.log("ERROR: Cannot parse CAPTCHA image URL from the page. Changing Tor circuit.", level=LogLevel.ERROR)
                        self.stats["all"] += 1
                        self.stats["net"] += 1
                        solver.stats(self.stats)
                        reload = True
                        continue

                    captcha_data = {}
                    for name in ("_token_", "timestamp", "salt", "hash", "captcha_type", "_do"):
                        captcha_data[name] = parse_single(r.text, r'name="' + re.escape(name) + r'" value="([^"]*)"')

                    # https://github.com/setnicka/ulozto-downloader/issues/82
                    captcha_image_url = urljoin("https:", captcha_image_url)

                    solver.log("Image URL obtained, trying to solve")
                    captcha_answer = solver.solve(captcha_image_url, stop_event)

                    captcha_data["captcha_value"] = captcha_answer

                    solver.log(f"CAPTCHA answer '{captcha_answer}' (timeout {self.conn_timeout})")

                    resp = s.post(self.captchaURL, data=captcha_data,
                                  headers=XML_HEADERS, timeout=self.conn_timeout)

                # generate result or break
                result = self._link_validation_stat(resp, solver.log)
                solver.stats(self.stats)
                # for noreload (bad captcha no need reload TOR)
                reload = result[1]
                if result[0]:
                    dlink = resp.json()["slowDownloadLink"]
                    # cache link here
                    self.linkCache.add(dlink)
                    self.numTorLinks += 1
                    yield dlink
                elif self.isDirectDownload:
                    solver.log("Direct download does no seem to work, trying with captcha resolution instead...")
                    self.isDirectDownload = False

            except requests.exceptions.ConnectionError:
                self._error_net_stat(
                    "Connection error, try new TOR session.", solver.log)
            except requests.exceptions.ChunkedEncodingError:
                self._error_net_stat(
                    "Error while communicating over Tor, try new TOR session", solver.log)
            except requests.exceptions.ReadTimeout:
                self._error_net_stat(
                    "ReadTimeout error, try new TOR session.", solver.log)

            solver.stats(self.stats)

    def enter_password(self, session):

        if not self.password and not self.frontend.supports_prompt:
            raise ValueError("The file requires a password. Provide it by re-running with '--password <password>'.")

        while True:
            if not self.password:
                self.password = self.frontend.prompt("File is password-protected, enter the password: ", level=LogLevel.WARNING)

            r = session.post(
                self.url,
                data={
                    "password": self.password,
                    "password_send": "Odeslat",
                    "_do": "passwordProtectedForm-submit",
                }
            )

            # Accept the password and store (auth) cookies
            if r.status_code == 200:
                print("Password accepted.")
                return r

            # Wrong password - retry when using frontend with prompt
            if self.frontend.supports_prompt:
                self.frontend.main_log("Wrong password, try again", level=LogLevel.ERROR)
                self.password = None
                continue

            raise ValueError("Wrong password")



from typing import Callable

import stem.process
import stem.control
from uldlib.utils import get_available_port

sockPort = get_available_port(9050)
controlPort = get_available_port(9051, skip=[sockPort])

TOR_CONFIG = {
    'SocksPort': f"{sockPort}",
    "ControlPort": f"{controlPort}",
    'SocksListenAddress': '127.0.0.1',
    'SocksPolicy': 'accept 127.0.0.1',
    'CookieAuthentication': '1',
    'Log': [
        'NOTICE stdout',
        'ERR file ./error.log'
    ],
}


class TorRunner:
    """
    A class that manages running and stopping a Tor process.
    """

    def __init__(self, temp_dir: str, log_func: Callable) -> None:
        """
        Initializes a TorRunner instance with a given data directory and log function.

        Args:
            temp_dir (str): the directory where the temporary data will be stored.
            log_func (Callable): a function that will be called to log messages.
        """
        self.tor_process = None
        self.log_func = log_func
        self.temp_dir = temp_dir
        self.proxies = {
            'http': f'socks5://127.0.0.1:{TOR_CONFIG.get("SocksPort")}',
            'https': f'socks5://127.0.0.1:{TOR_CONFIG.get("SocksPort")}'
        }

    def start(self) -> None:
        """
        Starts the Tor process with the given configuration.
        """
        try:
            self.tor_process = stem.process.launch_tor_with_config(config=TOR_CONFIG)
            self.log_func("TOR started")
        except Exception as e:
            self.log_func(f"Unable to start TOR: {e}")
            raise

    def launch(self) -> None:
        """
        Launches the Tor process if it has not been started.
        """
        if not self.tor_process:
            self.start()

    @staticmethod
    def reload() -> None:
        """
        Reloads the Tor process with a new circuit.
        """
        CONTROL_PORT = int(TOR_CONFIG.get("ControlPort"))
        with stem.control.Controller.from_port(port=CONTROL_PORT) as controller:
            controller.authenticate()
            controller.signal(stem.Signal.NEWNYM)

    def stop(self) -> None:
        """
        Stops the Tor process if running.
        """
        if self.tor_process:
            self.tor_process.kill()



import argparse
import importlib.util
import signal
import os
import sys
from os import path
from uldlib import downloader, captcha, __version__, __path__, const
from uldlib.frontend import ConsoleFrontend, JSONFrontend
from uldlib import utils
from uldlib.torrunner import TorRunner
from uldlib.utils import LogLevel

# TODO Automatic find all types implementing Frontend and put into this dict
available_frontends = {"console": ConsoleFrontend, "JSON": JSONFrontend}


def run():
    parser = argparse.ArgumentParser(
        description='Download file from Uloz.to using multiple parallel downloads.',
        usage=sys.argv[0]+" [options] URL...",
        add_help=False,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        'urls', metavar='URL', nargs="+", type=str,
        help="URL from Uloz.to (tip: enter in 'quotes' because the URL contains ! sign). Multiple URLs could be specified, they will be downloaded sequentially.")

    g_main = parser.add_argument_group("Main options")
    g_main.add_argument(
        '--parts', metavar='N', type=int, default=20,
        help='Number of parts that will be downloaded in parallel')
    g_main.add_argument(
        '--password', metavar='P', type=str, default="",
        help='Optional password if the file is password-protected')

    g_main.add_argument(
        '--output', metavar='DIRECTORY', type=str, default="./",
        help='Directory or full path including file name where output file will be saved')
    g_main.add_argument(
        '--temp', metavar='DIRECTORY', type=str, default="./",
        help='Directory where temporary files (.ucache, .udown, Tor data directory) will be created')
    g_main.add_argument(
        '-y', '--yes', default=False, action="store_true",
        help='Overwrite files without asking')

    g_log = parser.add_argument_group("Display and logging options")
    g_log.add_argument(
        '--parts-progress', default=False, action='store_true',
        help='Show progress of parts while being downloaded')
    g_log.add_argument(
        '--log', metavar='LOGFILE', type=str, default="",
        help="Enable logging to given file")
    g_log.add_argument(
        '--frontend', type=str, default="console", choices=available_frontends.keys(),
        help="Select frontend: 'console' - text user interface for humans, 'JSON' - output for scripts")

    g_captcha = parser.add_argument_group("CAPTCHA solving related options")
    g_captcha.add_argument(
        '--auto-captcha', default=False, action="store_true",
        help='Try to solve CAPTCHAs automatically using TensorFlow')
    g_captcha.add_argument(
        '--manual-captcha', default=False, action="store_true",
        help='Solve CAPTCHAs by manual input')
    g_captcha.add_argument(
        '--conn-timeout', metavar='SEC', default=const.DEFAULT_CONN_TIMEOUT, type=int,
        help='Set connection timeout for TOR sessions in seconds')

    g_other = parser.add_argument_group("Other options")
    g_other.add_argument('--version', action='version', version=__version__)
    g_other.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    args = parser.parse_args()

    # Use user chosen frontend
    frontend = available_frontends[args.frontend](show_parts=args.parts_progress, logfile=args.log)

    tfull_available = importlib.util.find_spec('tensorflow') and importlib.util.find_spec('tensorflow.lite')
    tflite_available = importlib.util.find_spec('tflite_runtime')
    tkinter_available = importlib.util.find_spec('tkinter')

    # Autodetection
    if not args.auto_captcha and not args.manual_captcha:
        if tfull_available:
            frontend.main_log("[Autodetect] tensorflow.lite available, using --auto-captcha")
            args.auto_captcha = True
        elif tflite_available:
            frontend.main_log("[Autodetect] tflite_runtime available, using --auto-captcha")
            args.auto_captcha = True
        elif tkinter_available:
            frontend.main_log("[Autodetect] tkinter available, using --manual-captcha")
            args.manual_captcha = True
        else:
            frontend.main_log(
                "[Autodetect] WARNING: No tensorflow.lite or tflite_runtime and no tkinter available, cannot solve CAPTCHA (only direct download available)",
                level=LogLevel.WARNING
            )

    if args.auto_captcha:
        if not (tfull_available or tflite_available):
            frontend.main_log('ERROR: --auto-captcha used but neither tensorflow.lite nor tflite_runtime are available', level=LogLevel.ERROR)
            sys.exit(1)

        model_path = path.join(__path__[0], const.MODEL_FILENAME)
        solver = captcha.AutoReadCaptcha(model_path, const.MODEL_DOWNLOAD_URL, frontend)
    elif args.manual_captcha:
        if not tkinter_available:
            frontend.main_log('ERROR: --manual-captcha used but tkinter not available', level=LogLevel.ERROR)
            sys.exit(1)

        solver = captcha.ManualInput(frontend)
    else:
        solver = captcha.Dummy(frontend)

    # enables ansi escape characters in terminal on Windows
    if os.name == 'nt':
        os.system("")

    tor = TorRunner(args.temp, frontend.tor_log)
    d = downloader.Downloader(tor, frontend, solver)

    # Register sigint handler
    def sigint_handler(sig, frame):
        if d.terminating:
            return  # Already terminating
        d.terminate()
        tor.stop()
        frontend.main_log('Program terminated.')
        sys.exit(1)

    signal.signal(signal.SIGINT, sigint_handler)

    try:
        for url in args.urls:
            d.download(url, args.parts, args.password, args.output, args.temp, args.yes, args.conn_timeout)
            # do clean only on successful download (no exception)
            d.clean()
    except utils.DownloaderStopped:
        pass
    except utils.DownloaderError as e:
        frontend.main_log(str(e), level=LogLevel.ERROR)
    finally:
        d.terminate()
        tor.stop()



import threading
import time

from datetime import timedelta
from typing import Tuple

from uldlib.utils import LogLevel
from uldlib.segfile import SegFileWriter


class DownloadPart:
    id: int
    writer: SegFileWriter
    download_url: str

    success: bool = False
    exception: Exception = None

    # Lock and protected variables (used by status thread)
    lock: threading.Lock
    started: bool
    completed: bool
    error: bool
    warning: bool
    status: str
    start_time: float
    completion_time: float
    size: int
    d_now: int
    d_total: int

    def __init__(self, writer: SegFileWriter):
        self.writer = writer
        self.id = writer.id
        self.lock = threading.Lock()

        # Init empty status
        self.started = False
        self.completed = False
        self.error = False
        self.status = ""
        self.size = writer.size
        self.d_now = 0
        self.d_total = writer.written

    def set_status(self, status: str, error: bool = False, warning: bool = False):
        self.lock.acquire()
        self.status = status
        self.error = error
        self.warning = warning
        self.lock.release()

    def get_frontend_status(self) -> Tuple[str, LogLevel, int]:
        """
        Returns status line for given part
        """

        level = LogLevel.INFO
        self.lock.acquire()
        downloaded = self.d_total

        if self.error:
            msg = self.status if self.status else "ERROR: Unknown error"
            level = LogLevel.ERROR
        elif self.warning:
            msg = self.status if self.status else "WARNING: Unknown warning"
            level = LogLevel.WARNING
        elif self.status and self.completed:
            msg = self.status
            level = LogLevel.SUCCESS
        elif self.status:
            msg = self.status
        elif self.completed:
            elapsed = self.completion_time - self.start_time
            speed = self.d_now / elapsed if elapsed > 0 else 0
            msg = "Successfully downloaded {}{} MB in {} (speed {} KB/s)".format(
                round(self.d_now / 1024**2, 2),
                "" if self.d_now == self.d_total else (
                    "/"+str(round(self.d_total / 1024**2, 2))
                ),
                str(timedelta(seconds=round(elapsed))),
                round(speed / 1024, 2)
            )
            level = LogLevel.SUCCESS
        else:
            elapsed = time.time() - self.start_time
            speed = self.d_now / elapsed if elapsed > 0 else 0
            # remaining time in seconds:
            remaining = (self.size - self.d_total) / speed if speed > 0 else 0

            msg = "{:.2f}%\t{:.2f}/{:.2f} MB\tspeed: {:.2f} KB/s\telapsed: {}\tremaining: {}".format(
                round(self.d_total / self.size * 100, 2),
                round(self.d_total / 1024**2, 2), round(self.size / 1024**2, 2),
                round(speed / 1024, 2),
                str(timedelta(seconds=round(elapsed))),
                str(timedelta(seconds=round(remaining))),
            )

        self.lock.release()
        return (msg, level, downloaded)



import os
from time import time
from typing import List
from urllib.parse import parse_qs

from .const import CACHEPOSTFIX


class LinkCache:
    """
    A class for caching download links.

    Attributes:
        filename (str): The name of the file where the links will be stored.
        shorten_validity (int, optional): number of seconds of witch shorten the validity of given link.

    Methods:
        delete_cache_file: Deletes the cache file if it exists.
        add: Adds a new link to the cache.
        get_all_valid_links: Returns all valid links from the cache.
        _is_link_valid: Determines whether a link is valid based on its timestamp query parameter.
        _get_cache_content: Returns the content of the cache file.
    """

    def __init__(self, filename: str, shorten_validity: int = 5):
        """
        Initializes a new instance of the LinkCache class.

        Args:
            filename (str): The name of the file where the links will be stored.
            shorten_validity (int, optional): number of seconds of witch shorten the validity of given link.
        """
        self.filename = filename
        self.cache_file = self.filename + CACHEPOSTFIX
        self.shorten_validity = shorten_validity

    def delete_cache_file(self) -> None:
        """
        Deletes the cache file if it exists.
        """
        if os.path.exists(self.cache_file):
            os.remove(self.cache_file)

    def add(self, link: str) -> None:
        """
        Adds a new link to the cache.
        """
        with open(self.cache_file, 'a') as cache:
            cache.write(f"{link}\n")

    def get_all_valid_links(self) -> List[str]:
        """
        Returns all valid links from the cache.
        """
        cache_content = self._get_cache_content()
        return [link for link in cache_content if self._is_link_valid(link)]

    def _is_link_valid(self, link: str) -> bool:
        """
        Determines whether a link is valid based on its timestamp query parameter.
        """
        query_string = parse_qs(link, separator=';')
        if not query_string.get("tm"):
            # link does not contain 'tm' query parameter
            # therefore is not valid
            return False
        link_timestamp = int(query_string.get("tm")[0])
        time_now = int(time())
        # flag link as invalid {shorten_validity} second before it actually expires
        return time_now < (link_timestamp - self.shorten_validity)

    def _get_cache_content(self) -> List[str]:
        """
        Returns the content of the cache file.
        """
        if not os.path.exists(self.cache_file):
            return []
        with open(self.cache_file, 'r') as cache:
            return cache.readlines()



#!/usr/bin/env python3

from distutils.core import setup
import setuptools  # noqa

from uldlib import __version__

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

setup(
    name='ulozto-downloader',
    version=__version__,
    license='MIT',
    description='Uloz.to quick multiple sessions downloader.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Jiří Setnička and Vlado Driver',
    author_email='setnicka@seznam.cz',
    url='https://github.com/setnicka/ulozto-downloader',
    package_data={'': ['model.tflite']},
    install_requires=[
        'requests',
        'Pillow',
        'ansicolors',
        'numpy',
        'pysocks',
        'stem',
    ],
    extras_require={
        "auto-captcha": [
            "tflite-runtime;python_version < '3.10'",
            "tensorflow;python_version >= '3.10'",
        ]
    },
    python_requires='>=3.7',
    packages=setuptools.find_packages(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    entry_points={
        'console_scripts': [
            'ulozto-downloader = uldlib.cmd:run',
            'uld = uldlib.cmd:run',
        ]
    }
)
